import { clone, setProperties, equals, toLookup, defaultHashFunction, arrayDiff } from "./Obj.js";
import { isArray, isNullOrUndefined, isNumber, isUndefined, isNotUndefined, isNotNullOrUndefined, isFunction, isObject, isNull } from "./Test.js";
import { assert, deprecate } from "./Util.js";
export { toLookup } from "./Obj.js";
export const difference = arrayDiff;
export const flatten = deprecate("Array.prototype.flatten", function flatten(src) {
    return src?.flat() ?? [];
});
export function reverse(array) {
    let left = null;
    let right = null;
    const length = isNullOrUndefined(array) ? 0 : array.length;
    for (left = 0; left < length / 2; left += 1) {
        right = length - 1 - left;
        const temporary = array[left];
        array[left] = array[right];
        array[right] = temporary;
    }
    return array;
}
export function concat(...arrs) {
    const result = Array.prototype.concat.apply([], arrs);
    return result;
}
export const slice = deprecate("Array.prototype.slice", function slice(src, pos = 0, count = Infinity) {
    return src?.slice(pos, pos + count) ?? [];
});
export const splice = deprecate("Array.prototype.splice", function (src, pos = 0, remove = Infinity, insert = []) {
    if (isNullOrUndefined(src)) {
        throw new Error("Unable to splice on null or undefined");
    }
    let srcLen = src.length;
    pos = Math.max(0, pos);
    pos = Math.min(pos, srcLen);
    remove = Math.max(0, remove);
    remove = Math.min(remove, srcLen - pos);
    return src.splice(pos, remove, ...insert);
});
export function append(arr, values) {
    let index = -1;
    const offset = arr.length;
    const length = isNullOrUndefined(values) ? 0 : values.length;
    while (++index < length) {
        arr[offset + index] = values[index];
    }
}
export function removeAt(arr, index) {
    return index >= 0 ? arr?.splice(index, 1)[0] : undefined;
}
export const indexOfElement = deprecate("Array.prototype.indexOf", function indexOfElement(src, el) {
    return src?.indexOf(el) ?? -1;
});
export function remove(arr, el) {
    const start = arr?.indexOf(el) ?? -1;
    removeAt(arr, start);
}
export function indexOf(src, fn) {
    let i = -1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    while (++i < len) {
        if (fn(src[i], i, src)) {
            return i;
        }
    }
    return -1;
}
export const find = deprecate("Array.prototype.find", function find(src, fn) {
    return src?.find(fn);
});
export function removeOneByFn(arr, fn) {
    const start = indexOf(arr, fn);
    removeAt(arr, start);
}
export function shallowCopy(src) {
    return src?.slice() ?? [];
}
export function shallowCopyInto(src, target) {
    let i = -1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    target.length = len;
    while (++i < len) {
        target[i] = src[i];
    }
}
export function shallowFill(src, target, at = 0) {
    let i = -1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    if (target.length < len + at) {
        target.length = len + at;
    }
    while (++i < len) {
        target[at + i] = src[i];
    }
}
export function deepCopy(src) {
    let i = -1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    const result = new Array(len);
    while (++i < len) {
        result[i] = (clone(src[i]));
    }
    return result;
}
export function deepCopyInto(src, target) {
    let i = -1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    target.length = len;
    while (++i < len) {
        target[i] = (clone(src[i]));
    }
}
export function deepFill(src, target, at = 0) {
    let i = -1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    if (target.length < len + at) {
        target.length = len + at;
    }
    while (++i < len) {
        target[at + i] = (clone(src[i]));
    }
}
export const filter = deprecate("Array.prototype.filter", function filter(src, fn) {
    return src?.filter(fn) ?? [];
});
export function filterInto(src, target, fn) {
    let i = -1;
    let j = 0;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    const space = target.length;
    while (++i < len) {
        const el = src[i];
        if (fn(el, i) === true) {
            if (j < space) {
                target[j++] = el;
            }
            else {
                ++j;
                target.push(el);
            }
        }
    }
    target.length = j;
}
export const map = deprecate("Array.prototype.map", function map(src, fn, startIndex = 0) {
    return src?.map(fn) ?? [];
});
export async function mapAsync(src, fn, inParallel = false) {
    let result;
    if (!inParallel) {
        let i = -1;
        const len = isNullOrUndefined(src) ? 0 : src.length;
        result = new Array(len);
        while (++i < len) {
            result[i] = await fn(src[i], i);
        }
    }
    else {
        result = await Promise.all(src.map(fn));
    }
    return result;
}
export function mapInto(src, target, fn, startIndex = 0) {
    let i = startIndex - 1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    target.length = len;
    while (++i < len) {
        target[i] = fn(src[i], i);
    }
}
export function reduce(src, fn, start, pos, to) {
    let acc = start;
    if (isNotNullOrUndefined(src)) {
        pos = Math.min(Math.max(0, isUndefined(pos) ? 0 : pos), src.length - 1);
        to = Math.min(Math.max(0, isUndefined(to) ? src.length - 1 : to), src.length - 1);
        let i = pos - 1;
        while (++i < to + 1) {
            acc = fn(acc, src[i]);
        }
    }
    return acc;
}
export function reduceUntil(src, fn, test, start, pos, to) {
    let acc = start;
    if (isNotNullOrUndefined(src)) {
        pos = Math.min(Math.max(0, isUndefined(pos) ? 0 : pos), src.length - 1);
        to = Math.min(Math.max(0, isUndefined(to) ? src.length - 1 : to), src.length - 1);
        let i = pos - 1;
        while (++i < to + 1 && !test(acc, src[i])) {
            acc = fn(acc, src[i]);
        }
    }
    return acc;
}
export function reverseReduce(src, fn, start) {
    let i = isNullOrUndefined(src) ? 0 : src.length;
    let acc = start;
    while (--i >= 0) {
        acc = fn(acc, src[i]);
    }
    return acc;
}
export function reverseReduceUntil(src, fn, test, start) {
    let i = isNullOrUndefined(src) ? 0 : src.length;
    let acc = start;
    while (--i >= 0 && !test(acc, src[i])) {
        acc = fn(acc, src[i]);
    }
    return acc;
}
export function forEach(src, fn, startIndex = 0) {
    let i = startIndex - 1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    while (++i < len) {
        fn(src[i], i);
    }
}
export async function forEachAsync(array, fn, inParallel = false) {
    if (!inParallel) {
        let i = -1;
        const len = isNullOrUndefined(array) ? 0 : array.length;
        while (++i < len) {
            await fn(array[i], i);
        }
    }
    else {
        await Promise.all(array.map(fn));
    }
}
export function forSome(src, filter, fn) {
    let i = -1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    while (++i < len) {
        const el = src[i];
        if (filter(el, i)) {
            fn(el, i);
        }
    }
}
export function until(src, fnOrTest, fn, startIndex) {
    let isCombined = isUndefined(fn) || isNumber(fn);
    startIndex = isCombined ? fn : startIndex;
    let i = isUndefined(startIndex) || startIndex < 0 ? -1 : startIndex - 1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    while (++i < len && (isCombined ?
        !fnOrTest(src[i], i) :
        !(fnOrTest(src[i], i) || (fn(src[i], i), false)))) {
    }
}
export function reverseForEach(src, fn) {
    let i = isNullOrUndefined(src) ? 0 : src.length;
    while (--i >= 0) {
        fn(src[i], i);
    }
}
export function reverseUntil(src, fnOrTest, fn) {
    let i = isNullOrUndefined(src) ? 0 : src.length;
    let combined = isUndefined(fn);
    while (--i >= 0 && (combined ? !fnOrTest(src[i], i) : !(fnOrTest(src[i], i) || (fn(src[i], i), false)))) {
    }
}
export const some = deprecate("Array.prototype.some", function some(src, fn) {
    return src?.some(fn) ?? false;
});
export function all(src, fn) {
    let result = true;
    let i = -1;
    const len = isNullOrUndefined(src) ? 0 : src.length;
    while (++i < len && (result = fn(src[i], i))) {
    }
    return result;
}
export function insertAt(src, pos, v) {
    if (isNullOrUndefined(src)) {
        throw new Error("Unable to insertAt on null or undefined");
    }
    if (pos === 0) {
        src.unshift(v);
    }
    else if (pos > 0) {
        let i = src.length;
        while (--i >= pos) {
            src[i + 1] = src[i];
        }
        src[i + 1] = v;
    }
}
export function binarySearch(src, cmp, closest = false) {
    let lo = 0, hi = isNullOrUndefined(src) ? -1 : src.length - 1, mid, element;
    while (lo <= hi) {
        mid = ((lo + hi) >> 1);
        element = src[mid];
        let val = cmp(element);
        if (val < 0) {
            lo = mid + 1;
        }
        else if (val > 0) {
            hi = mid - 1;
        }
        else {
            return mid;
        }
    }
    return closest ? lo : -1;
}
export function create(length, populator) {
    if ((length || -1) < 0) {
        length = 0;
    }
    let arr = new Array(length);
    let i = -1;
    while (++i < length) {
        arr[i] = populator(i, arr);
    }
    return arr;
}
export function zip(a, b, x, y, z) {
    let c;
    let d;
    let defaultZipFn = (i, a, b) => [a, b];
    let fn = isNotUndefined(x) ?
        isFunction(x) ?
            x :
            (c = x, isNotUndefined(y) ?
                isFunction(y) ?
                    y :
                    (d = y, isNotUndefined(z) ?
                        z :
                        defaultZipFn) :
                defaultZipFn) :
        defaultZipFn;
    let i = -1;
    let hasC = isNotUndefined(c);
    let hasD = isNotUndefined(d);
    let max = Math.min(a.length, b.length, hasC ? c.length : Number.POSITIVE_INFINITY, hasD ? d.length : Number.POSITIVE_INFINITY);
    let u;
    let result = [];
    if (hasC && hasD) {
        while (++i < max && (u = fn(i, a[i], b[i], c[i], d[i])) !== undefined) {
            result.push(u);
        }
    }
    else if (hasC) {
        while (++i < max && (u = fn(i, a[i], b[i], c[i])) !== undefined) {
            result.push(u);
        }
    }
    else {
        while (++i < max && (u = fn(i, a[i], b[i])) !== undefined) {
            result.push(u);
        }
    }
    return result;
}
export function unzip(arr, fn = (u, i, out) => (out[0] = u[0], out[1] = u[1], out)) {
    let i = -1;
    let len = arr.length;
    let split = [undefined, undefined];
    let result = [
        new Array(),
        new Array()
    ];
    while (++i < len && (split = fn(arr[i], i, split))) {
        result[0].push(split[0]);
        result[1].push(split[1]);
    }
    return result;
}
export function pivot(arr) {
    assert(isArray(arr) && arr.length > 0 && isArray(arr[0]), "argument has to be a non-empty array of arrays");
    let result = [];
    let i = -1;
    let len = arr.length;
    let width = arr[0].length;
    while (++i < width) {
        result.push(create(len, (j) => {
            return arr[j][i];
        }));
    }
    return result;
}
export function deserialize(array, target, ...types) {
    let T;
    let passthroughT;
    T = types.shift();
    passthroughT = types;
    if (isNotUndefined(T)) {
        if (isNotUndefined(T.prototype.deserialize)) {
            mapInto(array, target, (el) => {
                let t = new T();
                return t.deserialize.apply(t, [el].concat(passthroughT));
            });
        }
        else {
            mapInto(array, target, (el, i) => {
                let newT = new T();
                setProperties(newT, el);
                return newT;
            });
        }
    }
    else {
        deepCopyInto(array, target);
    }
}
export function bucket(array, ...desciminators) {
    let i = -1;
    const len = isNullOrUndefined(array) ? 0 : array.length;
    const descLen = isNullOrUndefined(desciminators) ? 0 : desciminators.length;
    let result = create(descLen + 1, () => []);
    while (++i < len) {
        let j = -1;
        let match = false;
        while (++j < descLen && !match) {
            if (desciminators[j](array[i])) {
                result[j].push(array[i]);
                match = true;
            }
        }
        if (!match) {
            result[j].push(array[i]);
        }
    }
    return result;
}
export function split(array, isA) {
    return bucket(array, isA);
}
export function distinct(array, hashFn) {
    let i = -1;
    const len = isNullOrUndefined(array) ? 0 : array.length;
    let result = [];
    let lookup = Object.create(null);
    let nullFound = false;
    let undefinedFound = false;
    let objects = [];
    let hasHashFn = !!hashFn;
    if (len > 0) {
        if (hasHashFn) {
            while (++i < len) {
                let el = array[i];
                let key = hashFn(el);
                if (!lookup[key]) {
                    result.push(el);
                    lookup[key] = true;
                }
            }
        }
        else {
            while (++i < len) {
                let el = array[i];
                if (isNull(el)) {
                    if (!nullFound) {
                        result.push(el);
                        nullFound = true;
                    }
                }
                else if (isUndefined(el)) {
                    if (!undefinedFound) {
                        result.push(el);
                        undefinedFound = true;
                    }
                }
                else if (isObject(el)) {
                    if (!objects.find((obj) => equals(obj, el))) {
                        result.push(el);
                        objects.push(el);
                    }
                }
                else {
                    let key = el.toString();
                    if (!lookup[key]) {
                        result.push(el);
                        lookup[key] = true;
                    }
                }
            }
        }
    }
    return result;
}
function shortLong(a, b) {
    return a.length < b.length ? [a, b] : [b, a];
}
export function intersect(a, b, hashFn = defaultHashFunction) {
    const result = [];
    const [_short, _long] = shortLong(a, b);
    const longLookup = toLookup(_long, hashFn);
    _short.forEach((el) => {
        if (longLookup[hashFn(el)]) {
            result.push(el);
        }
    });
    return result;
}
export function union(a, b, hashFn = defaultHashFunction) {
    const result = a.concat(b);
    return distinct(result, hashFn);
}
export function subtract(a, b, hashFn = defaultHashFunction) {
    const result = [];
    const bLookup = toLookup(b, hashFn);
    a.forEach((el) => {
        if (!bLookup[hashFn(el)]) {
            result.push(el);
        }
    });
    return result;
}
//# sourceMappingURL=Arr.js.map