import { getDate, assert } from './Util.js';
import { clone, mixin, transform } from './Obj.js';
function findCookie(key) {
    return document.cookie.split(';')
        .map((cookie) => cookie.trim())
        .find((cookie) => cookie.indexOf(`${key}=`) === 0);
}
export function getCookie(key) {
    let cookie = findCookie(key);
    return cookie ? cookie.trim().split('=')[1] : undefined;
}
export function setCookie(key, value, expires, path, sameSite = 'Lax', requireSSL = false) {
    document.cookie = `${key}=${value}; expires=${expires.toUTCString()}${path ? '; path=' + path : ''};SameSite=${sameSite ?? 'Lax'};${requireSSL || sameSite === 'None' ? ' Partitioned; Secure;' : ''}`;
}
export function removeCookie(key, path, sameSite = 'Lax', requireSSL = false) {
    document.cookie = `${key}=${'null'}; expires=${new Date(0).toUTCString()}${path ? '; path=' + path : ''};SameSite=${sameSite ?? 'Lax'};${requireSSL || sameSite === 'None' ? ' Partitioned; Secure;' : ''}`;
}
export function parseAllCookies() {
    return document.cookie
        .split(';')
        .map((cookie) => cookie.trim())
        .reduce((p, c, i) => {
        let [key, value] = c.split('=');
        p[key] = value;
        return p;
    }, {});
}
export function getMonster(options) {
    return new CookieMonster(options);
}
class CookieMonster {
    constructor(options) {
        this._defaultOptions = {
            name: '',
            defaults: {},
            retainTime: '7d',
            path: '',
            localStorage: false,
            session: false,
            sameSite: 'Lax',
            requireSSL: false,
        };
        this._defaultsKeyIndexes = {};
        this._options = this._defaultOptions;
        this._cookies = {};
        this._options = { ...this._defaultOptions, ...options };
        assert(this._options.name !== '', 'The Cookie Monster must have a name');
        assert(Object.keys(this._options.defaults).length > 0, 'Empty defaults object makes no sense');
        this._defaultsKeyIndexes = Object.keys(this._options.defaults).reduce((p, c, i) => ((p[c] = i), p), {});
        this._cookies = clone(this._options.defaults);
        this.openJar();
    }
    setCookie(key, value) {
        if (typeof this._cookies[key] !== typeof value) {
            throw new Error(`Type mismatch. Expected ${typeof this._cookies[key]} but got ${typeof value}`);
        }
        this._cookies[key] = value;
        this.saveCookies();
    }
    getCookie(key) {
        return this._cookies[key];
    }
    eatCookie(key) {
        this._cookies[key] = this._options.defaults[key];
        this.saveCookies();
    }
    removeCookies() {
        if (this._options.localStorage) {
            if (this._options.session) {
                sessionStorage.removeItem(this._options.name);
            }
            else {
                localStorage.removeItem(this._options.name);
            }
        }
        else {
            removeCookie(this._options.name, this._options.path, this._options.sameSite, this._options.requireSSL);
        }
    }
    saveCookies() {
        if (this._options.localStorage) {
            if (this._options.session) {
                sessionStorage.setItem(this._options.name, this.makeRecipe(this._cookies));
            }
            else {
                localStorage.setItem(this._options.name, this.makeRecipe(this._cookies));
            }
        }
        else {
            let expires = getDate(this._options.retainTime);
            setCookie(this._options.name, this.makeRecipe(this._cookies), expires, this._options.path, this._options.sameSite, this._options.requireSSL);
        }
    }
    makeRecipe(cookie) {
        let essentials = transform(cookie, (acc, value, key) => {
            if (value !== this._options.defaults[key]) {
                ;
                acc[key] = value;
            }
        }, {});
        let recipe = this.deflateNames(essentials);
        return this.escape(JSON.stringify(recipe));
    }
    deflateNames(essentials) {
        let recipe = transform(essentials, (acc, value, key) => {
            if (key in this._defaultsKeyIndexes) {
                acc[this._defaultsKeyIndexes[key].toString()] = value;
            }
        }, {});
        return recipe;
    }
    inflateNames(recipe) {
        let keyLookup = Object.keys(this._options.defaults);
        let essentials = transform(recipe, (acc, value, key) => {
            let index = parseInt(key, 10);
            let isIndex = !isNaN(index) && index.toString() === key;
            if (isIndex) {
                ;
                acc[keyLookup[index]] = value;
            }
        }, {});
        return essentials;
    }
    bakeCookies(recipe) {
        return {
            ...this._options.defaults,
            ...this.inflateNames(JSON.parse(this.unescape(recipe))),
        };
    }
    escape(str) {
        return str.replace(/;/g, '¤s').replace(/=/g, '¤e');
    }
    unescape(str) {
        return str.replace(/¤s/g, ';').replace(/¤e/g, '=');
    }
    openJar() {
        let local_recipe = localStorage.getItem(this._options.name);
        let session_recipe = sessionStorage.getItem(this._options.name);
        let cookie_recipe = getCookie(this._options.name);
        let recipe = this._options.localStorage
            ? this._options.session
                ? session_recipe
                : local_recipe
            : cookie_recipe;
        this.vacuum();
        if (recipe) {
            let cookies = this.bakeCookies(recipe);
            this._cookies = mixin(clone(this._cookies), null, cookies);
            this.saveCookies();
        }
    }
    vacuum() {
        if (!(this._options.localStorage && !this._options.session))
            localStorage.removeItem(this._options.name);
        if (!(this._options.localStorage && this._options.session))
            sessionStorage.removeItem(this._options.name);
        if (this._options.localStorage && findCookie(this._options.name)) {
            removeCookie(this._options.name, this._options.path, this._options.sameSite, this._options.requireSSL);
        }
    }
}
//# sourceMappingURL=Cookie.js.map