import { isNotNullOrUndefined, isNotUndefined, isFunction } from "./Test.js";
import { debounce, AssertError, throttle, once as runOnlyOnce } from "./Util.js";
export function debounced(duration, options) {
    return function innerDecorator(target, key, descriptor) {
        return {
            configurable: true,
            enumerable: descriptor.enumerable,
            get: function getter() {
                Object.defineProperty(this, key, {
                    configurable: true,
                    enumerable: descriptor.enumerable,
                    value: debounce(descriptor.value, duration, options)
                });
                return this[key];
            }
        };
    };
}
export function throttled(duration, options) {
    return function innerDecorator(target, key, descriptor) {
        return {
            configurable: true,
            enumerable: descriptor.enumerable,
            get: function getter() {
                Object.defineProperty(this, key, {
                    configurable: true,
                    enumerable: descriptor.enumerable,
                    value: throttle(descriptor.value, duration, options)
                });
                return this[key];
            }
        };
    };
}
export function once(target, propertyKey, descriptor) {
    const orgFn = descriptor.value;
    let result;
    let first = true;
    descriptor.value = function (...args) {
        if (first) {
            result = orgFn.apply(this, args);
            first = false;
        }
        return result;
    };
    return descriptor;
}
export function deprecated(instead, message) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        const className = target.constructor.name;
        let localMessage = isNotNullOrUndefined(message) ?
            message.replace("{class}", className).replace("{name}", propertyKey).replace("{instead}", instead || "") :
            "Function {class}::{name} is deprecated".replace("{class}", className).replace("{name}", propertyKey) +
                (isNotNullOrUndefined(instead) ? " please use {instead} instead".replace("{instead}", instead) : "");
        const warn = runOnlyOnce(function () {
            console.warn(localMessage);
        });
        descriptor.value = function (...args) {
            warn();
            return orgFn.apply(this, args);
        };
        return descriptor;
    };
}
export function asserts(assertFn, result) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        const className = target.constructor.name;
        const fnName = propertyKey;
        descriptor.value = function (...args) {
            try {
                assertFn.apply(this, args);
            }
            catch (assertError) {
                if (isNotUndefined(result)) {
                    return isFunction(result) ?
                        result(assertError) :
                        result;
                }
                else {
                    throw new AssertError(`Assertion failed in ${className}::${fnName}: ${assertError.message}`);
                }
            }
            return orgFn.apply(this, args);
        };
        return descriptor;
    };
}
export function transform(decoration) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        descriptor.value = function (...args) {
            let newArgs = decoration.apply(this, [orgFn.name].concat(args));
            const result = orgFn.apply(this, newArgs);
            return result;
        };
        return descriptor;
    };
}
export function before(decoration) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        descriptor.value = function (...args) {
            decoration.apply(this, [orgFn.name].concat(args));
            const result = orgFn.apply(this, args);
            return result;
        };
        return descriptor;
    };
}
export function after(decoration) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        descriptor.value = function (...args) {
            const result = orgFn.apply(this, args);
            decoration.apply(this, [orgFn.name].concat(args));
            return result;
        };
        return descriptor;
    };
}
export function around(decoration) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        descriptor.value = function (...args) {
            let result;
            let callback = () => result = orgFn.apply(this, args);
            decoration.apply(this, [callback, orgFn.name].concat(args));
            return result;
        };
        return descriptor;
    };
}
export function provided(condition) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        descriptor.value = function (...args) {
            let result;
            if (condition.apply(this, [orgFn.name].concat(args))) {
                result = orgFn.apply(this, args);
            }
            return result;
        };
        return descriptor;
    };
}
export let async = function async(target, propertyKey, descriptor) {
    const orgFn = descriptor.value;
    descriptor.value = function (...args) {
        let result = new Promise((resolve, reject) => {
            let value;
            try {
                value = orgFn.apply(this, args);
                resolve(value);
            }
            catch (err) {
                reject(err);
            }
        });
        return result;
    };
    return descriptor;
};
async.before = function (decoration) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        descriptor.value = function (...args) {
            let result = new Promise((resolve, reject) => {
                const callback = () => {
                    orgFn.apply(this, args)
                        .then((value) => {
                        resolve(value);
                    })
                        .catch((reason) => {
                        reject(reason);
                    });
                };
                decoration.apply(this, [orgFn.name].concat(args))
                    .then((value) => {
                    callback();
                })
                    .catch((reason) => {
                    callback();
                });
            });
            return result;
        };
        return descriptor;
    };
};
async.after = function (decoration) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        descriptor.value = function (...args) {
            let result = new Promise((resolve, reject) => {
                const callback = (value, reason) => {
                    decoration.apply(this, [value, reason, orgFn.name].concat(args))
                        .then((value) => {
                        resolve(value);
                    })
                        .catch((reason) => {
                        reject(reason);
                    });
                };
                orgFn.apply(this, args)
                    .then((value) => {
                    callback(value, undefined);
                })
                    .catch((reason) => {
                    callback(undefined, reason);
                });
            });
            return result;
        };
        return descriptor;
    };
};
async.provided = function (async_predicate) {
    return function (target, propertyKey, descriptor) {
        const orgFn = descriptor.value;
        descriptor.value = function (...args) {
            let result = new Promise((resolve, reject) => {
                const callback = () => {
                    orgFn.apply(this, args)
                        .then((value) => {
                        resolve(value);
                    })
                        .catch((reason) => {
                        reject(reason);
                    });
                };
                async_predicate.apply(this, [orgFn.name].concat(args))
                    .then((value) => {
                    if (value) {
                        callback();
                    }
                    else {
                        reject(false);
                    }
                })
                    .catch((reason) => {
                    reject(reason);
                });
            });
            return result;
        };
        return descriptor;
    };
};
//# sourceMappingURL=Decorators.js.map