import { Global } from "./Global.js";
import { isNullOrUndefined, isString, isElement } from "./Test.js";
import { assert } from "./Util.js";
import { reverse, reduce } from "./Arr.js";
export var Sides;
(function (Sides) {
    Sides[Sides["Top"] = 0] = "Top";
    Sides[Sides["Bottom"] = 1] = "Bottom";
    Sides[Sides["Left"] = 2] = "Left";
    Sides[Sides["Right"] = 3] = "Right";
})(Sides || (Sides = {}));
class DomState {
}
DomState.Sides = Sides;
DomState._window = Global.window;
DomState._document = Global.window ? Global.window.document : undefined;
DomState._el = Global.window ? Global.window.document.createElement("div") : undefined;
DomState._parser = Global.window ? new DOMParser() : undefined;
DomState._template = Global.window ? Global.window.document.createElement("template") : undefined;
DomState._fragment = Global.window ? Global.window.document.createDocumentFragment() : undefined;
export function init(win) {
    Global.window = win;
    DomState._window = Global.window;
    DomState._document = DomState._window.document;
    DomState._el = DomState._document.createElement("div");
}
export function create(html, attr) {
    let result;
    result = fromHTML(html)[0];
    setAttr(result, attr);
    return result;
}
export function fromHTML(html) {
    let result;
    let usesTemplate = !!(DomState._template && DomState._template.content);
    let usesParser = !!(DomState._parser);
    html = html.trim();
    if (/^[a-zA-Z]+$/.test(html)) {
        result = [Global.window.document.createElement(html)];
    }
    else {
        if (usesTemplate) {
            let template = DomState._template;
            template.innerHTML = html;
            assert(!!template.content.firstChild, "Dom.create was unable to parse html");
            result = Array.from(template.content.children);
            clear(template.content);
        }
        else if (usesParser) {
            let parser = DomState._parser;
            let doc = parser.parseFromString(html, "text/html");
            assert(!!doc.body.firstChild, "Dom.create was unable to parse html");
            result = Array.from(doc.body.children);
        }
        else {
            DomState._el.innerHTML = html;
            assert(!!DomState._el.firstChild, "Dom.create was unable to parse html");
            result = Array.from(DomState._el.children);
            clear(DomState._el);
        }
    }
    return result;
}
export function resolve(target, root) {
    let result = isString(target) ?
        (root && root !== DomState._document && is(target, root) ?
            root :
            find(target, root)) :
        target;
    return result;
}
export function resolveAll(target, root) {
    let result = isString(target) ?
        (root && root !== DomState._document && is(target, root) ?
            [root].concat(findAll(target, root)) :
            findAll(target, root)) :
        (isElement(target) ?
            [target] :
            target);
    return result;
}
export function contains(target, root, includeRoot = false) {
    const rootNode = root || DomState._document;
    let targets = resolveAll(target, root);
    const result = reduce(targets, (acc, cur) => {
        if (acc && cur) {
            if (!includeRoot) {
                cur = cur.parentNode;
            }
            while (cur !== rootNode && cur !== null) {
                cur = cur.parentNode;
            }
        }
        return acc && cur === rootNode;
    }, true);
    return result;
}
export function outerHTML(el) {
    DomState._el.appendChild(el);
    const result = DomState._el.innerHTML;
    clear(DomState._el);
    return result;
}
export function setAttr(_el, attr) {
    let el;
    if (typeof (_el) === "string") {
        el = byId(_el);
    }
    else {
        el = _el;
    }
    let keys, i, k, styles, styleKeys, style;
    if (el !== null && attr !== undefined && typeof (attr) === "object") {
        keys = Object.keys(attr);
        for (i = 0; i < keys.length; i++) {
            if (keys[i] === "style") {
                styles = attr[keys[i]];
                styleKeys = Object.keys(styles);
                for (k = 0; k < styleKeys.length; k++) {
                    style = styles[styleKeys[k]];
                    if (typeof (style) === "string") {
                        el.style.setProperty(styleKeys[k], style);
                    }
                    else {
                        el.style.setProperty(styleKeys[k], style[0], style[1] ?? undefined);
                    }
                }
            }
            else if ((keys[i] === "classes" || keys[i] === "class" || keys[i] === "className") && attr[keys[i]] !== undefined) {
                if (attr[keys[i]].join) {
                    el.className = attr[keys[i]].join(" ");
                }
                else {
                    el.className = attr[keys[i]];
                }
            }
            else {
                el.setAttribute(keys[i], attr[keys[i]]);
            }
        }
    }
}
export function remove(element) {
    return isNullOrUndefined(element.parentNode) ? null : element.parentNode.removeChild(element);
}
export function replace(src, target) {
    let result = target;
    if (src.parentNode) {
        src.parentNode.replaceChild(target, src);
    }
    return result;
}
export function clear(element) {
    let i = element.childNodes.length;
    while (i--) {
        element.removeChild(element.childNodes[i]);
    }
}
export function get(id) {
    return byId(id);
}
export function byId(id) {
    let result = DomState._document.getElementById(id);
    if (result === null) {
        switch (id) {
            case "body":
                result = DomState._document.body;
                break;
            default:
                break;
        }
    }
    return result;
}
export function find(selector, root) {
    return (root || DomState._document).querySelector(selector);
}
export function findAll(selector, root) {
    return Array.from((root || DomState._document).querySelectorAll(selector));
}
export function children(root, selector) {
    const children = Array.from((root || DomState._document).children);
    return selector === undefined ? children : children.filter((el) => is(selector, el));
}
export function findParent(start, parent) {
    let result = start.parentElement;
    while (result) {
        if (is(parent, result)) {
            break;
        }
        result = result.parentElement;
    }
    return result;
}
export function position(el, x, y) {
    el.style.top = `${y}px`;
    el.style.left = `${x}px`;
}
export function getOffset(el) {
    const rect = el.getBoundingClientRect();
    const scrollLeft = window.pageXOffset || document.documentElement.scrollLeft;
    const scrollTop = window.pageYOffset || document.documentElement.scrollTop;
    return {
        left: (rect.left + scrollLeft) | 0,
        top: (rect.top + scrollTop) | 0,
        right: (rect.right + scrollTop) | 0,
        bottom: (rect.bottom + scrollLeft) | 0,
    };
}
export function is(selector, element) {
    let result = false;
    if (!selector || !element || selector === document) {
    }
    else if (selector instanceof Element) {
        result = selector === element;
    }
    else if (element.matches) {
        result = element.matches(selector);
    }
    else if (element.msMatchesSelector) {
        result = (element.msMatchesSelector)(selector);
    }
    else if (element.webkitMatchesSelector) {
        result = element.webkitMatchesSelector(selector);
    }
    else {
        assert(!(isElement(selector) || isString(element)), "Did you mix up the arguments for Dom::is() ?");
        throw new Error("Browser does not support matches");
    }
    return result;
}
export function setStylesExplicitly(element, ...styles) {
    const comp = DomState._window.getComputedStyle(element);
    for (let i = 0; i < styles.length; i++) {
        let style = styles[i];
        element.style[style] = comp[style];
    }
}
export function findParents(element) {
    let result = [];
    while (element.parentElement) {
        element = element.parentElement;
        result.push(element);
    }
    return result;
}
export function findClosestCommonParent(elements, startAtIndex = 0) {
    let result = document.body;
    if (elements.length === 1) {
        result = elements[0];
    }
    else if (elements.length > 1) {
        let parents = elements.map((el) => {
            let parents = findParents(el);
            assert(parents.length >= startAtIndex, "startAtIndex higher than element parent depth");
            if (startAtIndex > 0) {
                while (parents.length > 0 && startAtIndex > 0) {
                    parents.pop();
                    --startAtIndex;
                }
            }
            let result = reverse(parents);
            result.push(el);
            return result;
        });
        let minDepth = parents.reduce((p, c, i) => Math.min(p, c.length), Infinity);
        if (minDepth > 0) {
            let firstList = parents.shift();
            let found = false;
            let i = 0;
            while (!found && i < minDepth) {
                found = !parents.reduce((p, c) => p === true && c[i] === firstList[i], true);
                ++i;
            }
            result = !found ? firstList[i - 1] : firstList[i - 2];
        }
    }
    return result;
}
//# sourceMappingURL=Dom.js.map