import { areNotNullOrUndefined, isArray, isFunction, isObject, isNullOrUndefined } from "./Test.js";
export function destroy(obj) {
    if (obj.constructor && isFunction(obj.constructor.prototype.destroy)) {
        obj.destroy();
    }
    else {
        setNull(obj);
    }
}
export function wipe(obj) {
    if (isArray(obj)) {
        obj.length = 0;
    }
    else if (isObject(obj)) {
        const keys = Object.keys(obj);
        let i = -1;
        const len = keys.length;
        while (++i < len) {
            delete obj[keys[i]];
        }
    }
}
export function setNull(obj) {
    if (obj.constructor && isFunction(obj.constructor.prototype.clear)) {
        obj.clear();
    }
    else {
        const keys = Object.keys(obj);
        let key = null;
        let i = -1;
        const len = keys.length;
        while (++i < len) {
            key = keys[i];
            obj[key] = null;
        }
    }
}
export function isClassOf(a, b) {
    return areNotNullOrUndefined(a, b) && a instanceof b.constructor;
}
export function isSameClass(a, b) {
    return areNotNullOrUndefined(a, b) && a.constructor === b.constructor;
}
export function inherits(a, b) {
    return isClassOf(a, b) && !isSameClass(a, b);
}
export function equals(a, b) {
    let result = a === b;
    if (a !== b && (a instanceof Object) && isSameClass(a, b)) {
        if (isArray(a)) {
            const len = a.length;
            let i = 0;
            result = len === b.length;
            if (result) {
                for (; i < len; i += 1) {
                    result = equals(a[i], b[i]);
                    if (result === false) {
                        break;
                    }
                }
            }
        }
        else if (a.constructor && isFunction(a.constructor.prototype.equals)) {
            result = a.equals(b);
        }
        else {
            const aKeys = Object.keys(a);
            const bKeys = Object.keys(b);
            let sameLength = aKeys.length === bKeys.length;
            let key = null;
            result = sameLength;
            let i = -1;
            const len = aKeys.length;
            while (++i < len && result) {
                key = aKeys[i];
                result = equals(a[key], b[key]);
                if (!result) {
                    if (isFunction(a[key]) && isFunction(b[key])) {
                        result = true;
                    }
                }
            }
        }
    }
    return result;
}
export function isDifferent(a, b) {
    return !equals(a, b);
}
export function shallowCopy(obj) {
    const keys = Object.keys(obj);
    const result = {};
    let i = -1;
    const len = keys.length;
    while (++i < len) {
        const key = keys[i];
        result[key] = obj[key];
    }
    return result;
}
export function clone(obj) {
    let result;
    let isNullObject = !isNullOrUndefined(obj) && !(obj instanceof Object) && !obj.constructor;
    if (!(obj instanceof Object) && !isNullObject) {
        result = obj;
    }
    else if ((obj instanceof Object) && isFunction(obj.constructor.prototype.clone)) {
        result = obj.clone();
    }
    else if (isArray(obj)) {
        let i = -1;
        const len = obj.length;
        result = new Array(len);
        while (++i < len) {
            result[i] = (clone(obj[i]));
        }
    }
    else if (obj instanceof Date) {
        return new Date(obj.getTime());
    }
    else if (obj instanceof RegExp) {
        return new RegExp(obj);
    }
    else if (isFunction(obj)) {
        result = obj;
    }
    else {
        if (isNullObject) {
            result = Object.create(null);
        }
        else {
            result = new obj.constructor();
        }
        const keys = Object.keys(obj);
        let key = null;
        let i = -1;
        const len = keys.length;
        while (++i < len) {
            key = keys[i];
            result[key] = clone(obj[key]);
        }
    }
    return result;
}
export function cloneInto(src, target) {
    if (isArray(target)) {
        const arrS = src;
        const arrT = target;
        const len = arrS.length;
        arrT.length = len;
        let i = -1;
        while (++i < len) {
            if (arrS[i] instanceof Object) {
                cloneInto(arrS[i], arrT[i]);
            }
            else {
                arrT[i] = arrS[i];
            }
        }
    }
    else {
        const keys = Object.keys(src);
        let key = null;
        let i = -1;
        const len = keys.length;
        while (++i < len) {
            key = keys[i];
            const a = src[key];
            if (a instanceof Object) {
                let b = target[key];
                if (b === undefined || b === null) {
                    if (isArray(a)) {
                        b = target[key] = [];
                    }
                    else {
                        b = target[key] = {};
                    }
                }
                if (isDifferent(a, b)) {
                    cloneInto(a, b);
                }
            }
            else {
                target[key] = a;
            }
        }
    }
    return target;
}
export function mixin(target = {}, exclude, ...sources) {
    const result = target;
    const srcLength = sources?.length ?? 0;
    for (let i = 0; i < srcLength; i++) {
        let src = sources[i];
        if (isFunction(src)) {
            src = src.prototype;
        }
        if (src === undefined) {
            continue;
        }
        const keys = Object.keys(src);
        let key = null;
        const keysLength = keys.length;
        for (let k = 0; k < keysLength; k++) {
            key = keys[k];
            if (exclude && exclude.hasOwnProperty(key)) {
                continue;
            }
            if (isObject(target[key]) && isObject(src[key])) {
                mixin(target[key], exclude, src[key]);
            }
            else {
                target[key] = src[key] ?? target[key];
            }
        }
    }
    return result;
}
export function setProperties(target, values, mapping, limitToExisting = false) {
    const keys = Object.keys(values);
    let key;
    let i = -1;
    const len = keys.length;
    while (++i < len) {
        key = keys[i];
        if (mapping && key in mapping) {
            key = mapping[key];
        }
        if (!limitToExisting || key in target) {
            target[key] = values[keys[i]];
        }
    }
}
export function forEach(target, fn) {
    if (isArray(target)) {
        let i = -1;
        const len = target.length;
        while (++i < len && false !== fn(target[i], i.toString())) {
        }
    }
    else {
        const keys = Object.keys(target);
        let key;
        let i = -1;
        const len = keys.length;
        let run = true;
        while (run && ++i < len) {
            key = keys[i];
            run = false !== fn(target[key], key);
        }
    }
}
export function transform(target, fn, accumulator) {
    accumulator = accumulator || (isObject(target) ? Object.create(target) : {});
    forEach(target, (value, key) => {
        return fn(accumulator, value, key);
    });
    return accumulator;
}
export function defaultHashFunction(el) {
    return el === null
        ? "____null"
        : el === undefined
            ? "____undefined"
            : isObject(el)
                ? JSON.stringify(el)
                : el.toString();
}
export function toLookup(a, hashFn = defaultHashFunction) {
    return a.reduce((p, c) => {
        p[hashFn(c)] = true;
        return p;
    }, {});
}
export function arrayDiff(a, b, hashFn = defaultHashFunction) {
    const lookupb = toLookup(b);
    const lookupa = toLookup(a);
    const hashb = Object.keys(lookupb);
    const hasha = Object.keys(lookupa);
    return [
        a.filter((el, i) => !lookupb[hasha[i]]),
        b.filter((el, i) => !lookupa[hashb[i]]),
    ];
}
export function difference(target, base) {
    function changes(target, base) {
        return transform(target, function (result, value, key) {
            if (isDifferent(value, base[key])) {
                result[key] = (isObject(value) && isObject(base[key])) ? changes(value, base[key]) : value;
            }
        });
    }
    return changes(target, base);
}
//# sourceMappingURL=Obj.js.map