import { Global } from "./Global.js";
import { hasConsole, isNotUndefined, isNotNullOrUndefined, isUndefined } from "./Test.js";
class LoggableCounter {
    log() {
        console.log(`Counter ${this.name}: ${this.toString()}`);
    }
    constructor(name = "") {
        this.name = "";
        this._value = 0;
        this.name = name;
    }
    inc() {
        this._value++;
        return this;
    }
    reset() {
        this._value = 0;
        return this;
    }
    valueOf() {
        return this._value;
    }
    toString() {
        return this._value.toString();
    }
}
class UtilState {
}
UtilState._int = { 0: 0 };
UtilState._counter = { "": new LoggableCounter("") };
export function once(fn) {
    let result;
    let first = true;
    return function (...args) {
        if (first) {
            result = fn.apply(this, args);
            first = false;
        }
        return result;
    };
}
export function deprecate(instead, fn) {
    let localMessage = `Function ${fn.name} is deprecated, please use ${instead} instead`;
    const warn = once(function () {
        console.warn(localMessage);
    });
    return {
        [fn.name]: function (...args) {
            if (!Global.noDeprecationWarnings)
                warn();
            return fn(...args);
        }
    }[fn.name];
}
export function init(win) {
    if (win !== undefined) {
        Global.window = win;
    }
}
function stdTimezoneOffset(date) {
    let jan = new Date(date.getFullYear(), 0, 1);
    let jul = new Date(date.getFullYear(), 6, 1);
    return Math.max(jan.getTimezoneOffset(), jul.getTimezoneOffset());
}
export function isDaylightSavingTime(date) {
    return date.getTimezoneOffset() < stdTimezoneOffset(date);
}
export function getDate(delta = "", start) {
    const date = start ? new Date(start.getTime()) : new Date();
    const rel = delta.split(" ").join("");
    const [, sign, years, months, days, hours, mins, secs] = toArray(/([+-])?(?:(\d+)Y)?(?:(\d+)M)?(?:(\d+)d)?(?:(\d+)h)?(?:(\d+)m)?(?:(\d+)s)?/.exec(rel));
    const plus = sign !== "-";
    if (secs) {
        date.setSeconds(date.getSeconds() + (plus ? +parseInt(secs, 10) : -parseInt(secs, 10)));
    }
    if (mins) {
        date.setMinutes(date.getMinutes() + (plus ? +parseInt(mins, 10) : -parseInt(mins, 10)));
    }
    if (hours) {
        date.setHours(date.getHours() + (plus ? +parseInt(hours, 10) : -parseInt(hours, 10)));
    }
    if (days) {
        date.setDate(date.getDate() + (plus ? +parseInt(days, 10) : -parseInt(days, 10)));
    }
    if (months) {
        date.setMonth(date.getMonth() + (plus ? +parseInt(months, 10) : -parseInt(months, 10)));
    }
    if (years) {
        date.setFullYear(date.getFullYear() + (plus ? +parseInt(years, 10) : -parseInt(years, 10)));
    }
    return date;
}
export function newUUID() {
    let d = new Date().getTime();
    let hrTime = [0, 0];
    d += typeof (performance) !== "undefined"
        ? performance.now()
        : (hrTime = process.hrtime(), hrTime[0] * 1000 + (hrTime[1] / 1e6));
    return "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g, function (c) {
        const r = (d + Math.random() * 16) % 16 | 0;
        d = Math.floor(d / 16);
        return (c === "x" ? r : (r & 0x3 | 0x8)).toString(16);
    });
}
export function newInt(key = 0) {
    if (UtilState._int[key] === undefined) {
        UtilState._int[key] = 0;
    }
    return UtilState._int[key]++;
}
export function counter(key = "") {
    if (UtilState._counter[key] === undefined) {
        UtilState._counter[key] = new LoggableCounter(key.toString());
    }
    return UtilState._counter[key];
}
export function count(key = "") {
    return counter(key).inc();
}
export function pipeOut(log, warn, error, catchDefault = false) {
    if (hasConsole()) {
        let preventDefaultObj = catchDefault;
        if (isNotNullOrUndefined(log)) {
            proxyFn(Global.global.console, "log", function (superfn, ...args) {
                if (!(catchDefault || preventDefaultObj.log)) {
                    superfn.apply(this, args);
                }
                log.apply(this, args);
            });
        }
        if (isNotNullOrUndefined(warn)) {
            proxyFn(Global.global.console, "warn", function (superfn, ...args) {
                if (!(catchDefault || !preventDefaultObj.warn)) {
                    superfn.apply(this, args);
                }
                warn.apply(this, args);
            });
        }
        if (isNotNullOrUndefined(error)) {
            proxyFn(Global.global.console, "error", function (superfn, ...args) {
                if (!(catchDefault || !preventDefaultObj.error)) {
                    superfn.apply(this, args);
                }
                error.apply(this, args);
            });
        }
    }
}
export class AssertError extends Error {
    constructor(m) {
        super(m);
        Object.setPrototypeOf(this, AssertError.prototype);
    }
}
export function assert(assertion, message = "", noThrow = false) {
    let result = true;
    if (!assertion) {
        if (noThrow) {
            console.error("Assertion failed: " + message);
        }
        else {
            throw new AssertError(message);
        }
        result = false;
    }
    return result;
}
export function proxyFn(objOrClass, fnName, proxyFn) {
    objOrClass = isNotUndefined(objOrClass.prototype) ? objOrClass.prototype : objOrClass;
    const fn = objOrClass[fnName];
    const _superFn = function (...args) {
        if (args.length !== 0) {
            return fn.apply(this || objOrClass, args);
        }
        else {
            return fn.call(this || objOrClass);
        }
    };
    objOrClass[fnName] = proxyFn.bind(objOrClass, _superFn);
}
export function loop(count, fn) {
    let i = -1;
    let run = true;
    while (run && ++i < count) {
        run = fn(i) !== true;
    }
}
export function toArray(arr) {
    return Array.from(arr);
}
export const DEFAULT_DURATION = 100;
export function debounce(method, duration = DEFAULT_DURATION, options) {
    let timeoutHandle = null;
    let leading = isUndefined(options) || isUndefined(options.leading) ? false : options.leading;
    let executed = false;
    let result = undefined;
    let resolve;
    let reject;
    let wrapper = function (...args) {
        if (!leading) {
            result = result || new Promise((_resolve, _reject) => {
                resolve = _resolve;
                reject = _reject;
            });
        }
        if (timeoutHandle === null) {
            if (leading) {
                executed = true;
                result = method.apply(this, args);
            }
        }
        wrapper.resetTimer();
        timeoutHandle = setTimeout(() => {
            timeoutHandle = null;
            if (!executed) {
                let value = method.apply(this, args);
                if (isNotNullOrUndefined(value) && value.hasOwnProperty("then")) {
                    value.then((v) => {
                        resolve(v);
                    });
                }
                else {
                    resolve(value);
                }
            }
            executed = false;
            result = undefined;
        }, duration);
        return result;
    };
    wrapper.resetTimer = function () {
        if (timeoutHandle) {
            clearTimeout(timeoutHandle);
            timeoutHandle = null;
        }
    };
    return wrapper;
}
export function throttle(method, duration = DEFAULT_DURATION, options) {
    let timeoutHandle = null;
    let leading = isUndefined(options) || isUndefined(options.leading) ? true : options.leading;
    let trailing = !leading || isUndefined(options) || isUndefined(options.trailing) ? true : options.trailing;
    let result;
    let lastContext;
    let lastArgs;
    let wrapper = function (...args) {
        lastArgs = args;
        lastContext = this;
        if (timeoutHandle === null) {
            if (leading) {
                result = method.apply(lastContext, lastArgs);
            }
            if (!trailing) {
                lastContext = null;
                lastArgs = [];
            }
            timeoutHandle = setTimeout(() => {
                timeoutHandle = null;
                if (trailing) {
                    method.apply(lastContext, lastArgs);
                    lastContext = null;
                    lastArgs = [];
                }
            }, duration);
        }
        return result;
    };
    return wrapper;
}
//# sourceMappingURL=Util.js.map